#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2025, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Entry point into the IDMS Linux installer VM functionality from the commandline."""

import argparse
import logging
import signal
import sys
from types import FrameType
from typing import Any

from ..exceptions import IliRuntimeError, IliUsageError
from ..filesystem import IliPath
from ..version import __version__
from .logging import setup_logging
from .vm import (
    DEFAULT_VM_CONNECT_TIMEOUT,
    DEFAULT_VM_START_TIMEOUT,
    IliVM,
    IliVMNetworkDeviceBridge,
    IliVMScript,
    IliVMScriptHost,
    IliVMScriptVM,
    IliVMState,
)

__all__ = ["run_ilivm", "run_ilivm_entrypoint"]


def get_argparser(prog: str | None) -> argparse.ArgumentParser:
    """Return an instance of argparse.ArgumentParser."""

    print(f"IDMS Linux Installer VM v{__version__} - Copyright © 2019-2025, AllWorldIT.\n", file=sys.stderr)  # noqa: T201

    argparser = argparse.ArgumentParser(prog=prog, add_help=False)

    optional_group = argparser.add_argument_group("Optional arguments")
    optional_group.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    optional_group.add_argument(
        "-w", "--work-dir", nargs=1, default=None, help="Working directory, by default TMPDIR or /tmp is used."
    )
    optional_group.add_argument(
        "--wait-poweroff",
        nargs="?",
        default=None,
        type=int,
        help="Do not stop the VM after running the scripts and wait X seconds for poweroff, by default the VM is terminated. "
        "0 = no limit.",
    )
    optional_group.add_argument("-q", "--quiet", action="store_true", help="Be more quiet.")

    return argparser


def run_ilivm(argv: list[str], prog: str = "ilivm") -> None:  # noqa: C901,PLR0912,PLR0915
    """Entry point for execution from the commandline."""

    def handle_signal(signum: int, frame: FrameType | None) -> None:  # noqa: ARG001
        """Handle signals."""
        logging.error("[HOST] *** Received signal %s (%s) ***", signum, signal.Signals(signum).name)
        vm.stop()

    our_dir = IliPath(__file__).parent

    # Grab our arg parser
    argparser = get_argparser(prog)

    argparser.register("type", "IliVMScriptHost", lambda s: IliVMScriptHost(s))
    argparser.register("type", "IliVMScriptVM", lambda s: IliVMScriptVM(s))

    argparser.add_argument("--live-image", nargs=1, help="Path to the live image to use.")
    argparser.add_argument("--vm-name", nargs=1, help="Name of the VM to create.")
    argparser.add_argument(
        "--vm-start-timeout",
        nargs=1,
        type=int,
        default=DEFAULT_VM_START_TIMEOUT,
        help=f"Timeout in seconds to wait for VM to start (default: {DEFAULT_VM_START_TIMEOUT}).",
    )
    argparser.add_argument(
        "--vm-connect-timeout",
        nargs=1,
        type=int,
        default=DEFAULT_VM_CONNECT_TIMEOUT,
        help=f"Timeout in seconds to wait for VM to start (default: {DEFAULT_VM_CONNECT_TIMEOUT}).",
    )
    argparser.add_argument(
        "--copy-to-vm", action="append", help="File/paths to copy from the HOST to the VM in the format of SOURCE:DEST."
    )
    argparser.add_argument(
        "--copy-from-vm", action="append", help="File/paths to copy from the VM to the HOST in the format of SOURCE:DEST."
    )

    argparser.add_argument(
        "--vm-memory",
        nargs=1,
        type=int,
        help="Amount of RAM to allocate to the VM in MB. Default is 4096.",
    )
    argparser.add_argument(
        "--vm-cores",
        nargs=1,
        type=int,
        help="Number of CPU cores to allocate to the VM. Default is 1.",
    )

    argparser.add_argument(
        "--add-disk",
        action="append",
        help="Add a disk to the VM, either a path or an integer size in GB. "
        "Also supports adding ',model=MODEL' and ',serial=SERIAL' to the end of the disk string.",
    )

    argparser.add_argument(
        "--host-fs",
        nargs=1,
        help="Mount a host filesystem in the VM, if the path is specified it is used else a temporary path is created.",
    )

    argparser.add_argument(
        "--kernel",
        nargs=1,
        choices=["linux", "linux-lts"],
        default=["linux-lts"],
        help="Kernel to use, 'linux' or 'linux-lts'. Default is 'linux'.",
    )
    argparser.add_argument(
        "--extra-kernel-args",
        nargs=1,
        help="Extra kernel arguments to pass to the kernel.",
    )
    argparser.add_argument(
        "--network-bridge", nargs=1, help="Network bridge to use, by default the VM will use host-only networking."
    )
    argparser.add_argument(
        "--no-oem-autorun",
        action="store_true",
        help="Do not add the OEM autorun script.",
    )
    argparser.add_argument(
        "--add-oem-file",
        action="append",
        help="Add a file to the OEM media",
    )
    argparser.add_argument(
        "--run-vm-script",
        dest="run_scripts",
        metavar="SCRIPT",
        type="IliVMScriptVM",
        action="append",
        help="Script to run in VM once booted. This will be run in the order it is specified together with host scripts.",
    )
    argparser.add_argument(
        "--run-host-script",
        dest="run_scripts",
        metavar="SCRIPT",
        type="IliVMScriptHost",
        action="append",
        help="Script to run on host while VM is running. This will be run in the order it is specified together with VM scripts.",
    )
    argparser.add_argument(
        "--cache-unsafe",
        action="store_true",
        help="Use unsafe caching with no guarantee of data integrity. This is faster but can cause data corruption.",
    )

    # If no arguments were specified, display help
    if len(sys.argv) == 1:
        raise IliUsageError(argparser.format_help())

    # Parse args
    args = argparser.parse_args(argv)

    # Check if we were provided a VM name
    vm_name = None
    if args.vm_name:
        vm_name = args.vm_name[0]

    if args.quiet:
        setup_logging(vm_name=vm_name)
    else:
        setup_logging(level=logging.DEBUG, vm_name=vm_name)

    # Work out arguments to pass to the VM class
    vm_kwargs: dict[str, Any] = {}
    # Live image to boot off
    if args.live_image:
        vm_kwargs["live_image"] = IliPath(args.live_image[0]).expanduser()
    # Base working directory
    if args.work_dir:
        vm_kwargs["work_dir"] = IliPath(args.work_dir[0]).expanduser()

    # Create VM object
    vm = IliVM(**vm_kwargs)

    # Check if we're setting the number of CPU cores
    if args.vm_cores:
        vm.vm_cores = args.vm_cores[0]

    # Check if we're setting the amount of RAM
    if args.vm_memory:
        vm.vm_memory = args.vm_memory[0]

    # Check if we're using unsafe caching
    if args.cache_unsafe:
        vm.cache_unsafe = True

    # Check if we're setting the start timeout
    if args.vm_start_timeout:
        vm.start_timeout = args.vm_start_timeout

    # Work out files to copy to the VM
    copy_to_vm: list[tuple[IliPath, IliPath]] = []
    if args.copy_to_vm:
        # Loop and add each pair
        for copy_file in args.copy_to_vm:
            source, dest = copy_file.split(":")
            if not dest:
                raise ValueError(f"Invalid format for --add-to-vm (SOURCE:DEST): {copy_file}")
            source_full = IliPath(source).expanduser()
            if not source_full.exists():
                raise IliRuntimeError(f"Cannot find path to copy to VM: {source_full}")
            copy_to_vm.append((IliPath(source).expanduser(), IliPath(dest)))
    # Work out files to copy from the VM
    copy_fom_vm: list[tuple[IliPath, IliPath]] = []
    if args.copy_from_vm:
        # Loop and add each pair
        for copy_file in args.copy_from_vm:
            source, dest = copy_file.split(":")
            if not dest:
                raise ValueError(f"Invalid format for --get-from-vm (SOURCE:DEST): {copy_file}")
            copy_fom_vm.append((IliPath(source), IliPath(dest).expanduser()))

    # Load scripts to run
    scripts: list[IliVMScript] = []
    if args.run_scripts:
        scripts = args.run_scripts.copy()

    # Prepare VM
    vm.add_oem_file(
        our_dir.concatpath("data", "oem", "grub.cfg"),
        replaces={"@KERNEL@": args.kernel[0], "@EXTRA_KERNEL_ARGS@": args.extra_kernel_args[0] if args.extra_kernel_args else ""},
    )
    # Check if we're exlcuding the OEM autorun script
    if not args.no_oem_autorun:
        vm.add_oem_file(our_dir.concatpath("data", "oem", "autorun"))
    # Add additional OEM media files
    if args.add_oem_file:
        for oem_file in args.add_oem_file:
            vm.add_oem_file(IliPath(oem_file))

    # Check if we're adding a host filesystem
    if args.host_fs:
        vm.add_hostfs(IliPath(args.host_fs[0]).expanduser())

    # Add disks
    mac_from_str = ""
    if args.add_disk:
        # Loop with disk options, split off optional serial and model and add to VM
        for disk in args.add_disk:
            # Split disk components
            disk_components = disk.split(",")
            # Start with nothing
            disk_path = None
            disk_size = None
            disk_product = None
            disk_model = None
            disk_serial = None
            # Work out if its a size specification
            if disk_components[0].isdigit():
                # Size in GB
                disk_size = int(disk_components[0])
            else:
                disk_path = IliPath(disk_components[0]).expanduser()
            # Use the first path we find as the string to generate the MAC address
            if not mac_from_str and disk_path:
                mac_from_str = str(disk_path)
            # If not a size specification, its a path specification
            for component in disk_components[1:]:
                if component.startswith("product="):
                    disk_product = component.split("=", 1)[1]
                elif component.startswith("model="):
                    disk_model = component.split("=", 1)[1]
                elif component.startswith("serial="):
                    disk_serial = component.split("=", 1)[1]
                else:
                    raise IliRuntimeError(f"Invalid disk option: {component}")
            # Add disk to VM
            vm.add_disk(path=disk_path, size=disk_size, product=disk_product, model=disk_model, serial=disk_serial)

    # Check if we're setting the network bridge
    if args.network_bridge:
        network_bridge_device = IliVMNetworkDeviceBridge(mac_from_str=mac_from_str, bridge=args.network_bridge[0])
        vm.add_network_device(network_bridge_device)

    # Prepare VM
    vm.prepare()

    # We run the main process in a try/except block to catch any exceptions with a signal handler
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)
    # Deferred exception to raise later
    exception: Exception | None = None

    # Start VM
    vm.start()

    # Try copy files we need to the VM before running scripts
    for source, dest in copy_to_vm:
        try:
            vm.copy_to_vm(source, dest)
        except IliRuntimeError as exc:
            logging.error("[HOST]   - Failed to copy file '%s' to VM", source)  # noqa: TRY400
            logging.error("[HOST]     - %s", exc)  # noqa: TRY400
            break

    # Run scripts
    for script in scripts:
        try:
            script.run(vm)
        except IliRuntimeError as exc:
            logging.error("[HOST]   - Failed to run %s script '%s'", script.run_type.name, script)  # noqa: TRY400
            logging.error("[HOST]     - %s", exc)  # noqa: TRY400
            exception = exc
            break

    if vm.state == IliVMState.STOPPED:
        logging.info("[HOST] VM stopped unexpectedly")
        exception = IliRuntimeError("VM stopped unexpectedly")
    elif exception:
        logging.error("[HOST] One or more scripts failed to run")
    else:
        logging.info("[HOST] All scripts ran successfully")
        # Check if we need to copy files from VM
        if copy_fom_vm:
            logging.info("[HOST] Copying files from VM")
        for source, dest in copy_fom_vm:
            try:
                vm.copy_from_vm(source, dest)
            except IliRuntimeError as exc:
                logging.error("[HOST]   - Failed to copy file '%s' from VM", source)  # noqa: TRY400
                logging.error("[HOST]     - %s", exc)  # noqa: TRY400
                exception = exc
                break

    # Stop VM if its still running
    if vm.state == IliVMState.RUNNING:
        if args.wait_poweroff is None:
            vm.stop()
        else:
            poweroff_timeout = None if args.wait_poweroff == 0 else args.wait_poweroff
            vm.poweroff()
            vm.wait_for_vm_poweroff(poweroff_timeout)

    # Raise exception if it was deferred
    if exception:
        raise exception


def run_ilivm_entrypoint(prog: str = "ilivm") -> None:
    """Entry point for execution from the commandline."""
    try:
        run_ilivm(sys.argv[1:], prog)
    except IliUsageError as exc:
        print(str(exc), file=sys.stderr)  # noqa: T201
    # Catch installer exception
    except IliRuntimeError as exc:
        logging.error("!!!!! ERROR !!!!!")  # noqa: TRY400
        logging.error("Exception encountered: %s", exc)  # noqa: TRY400
        logging.error("!!!!! ERROR !!!!!")  # noqa: TRY400
        sys.exit(1)
