#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2025, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""IDMS Linux installer log formatter class."""

import logging
from typing import Any, ClassVar

__all__ = ["IliVMLogFormatter"]


class IliVMLogFormatter(logging.Formatter):
    """Custom log formatter with ANSI color support."""

    # Name of the VM
    _vm_name: str | None = None

    # ANSI escape codes for colors
    COLORS: ClassVar[dict[str, str]] = {
        "DEBUG": "\033[94m",  # Blue
        "INFO": "\033[92m",  # Green
        "WARNING": "\033[93m",  # Yellow
        "ERROR": "\033[91m",  # Red
        "CRITICAL": "\033[95m",  # Magenta
    }
    RESET = "\033[0m"  # Reset color
    BRIGHT_WHITE = "\033[1;37m"  # Bright white color

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize the formatter with default values."""
        # Extract vm_name from the kwards and don't pass it to the parent
        self._vm_name = kwargs.pop("vm_name", None)
        # Initialize the parent class
        super().__init__(*args, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record with ANSI color."""

        # Apply color based on log level
        log_color = self.COLORS.get(record.levelname, self.RESET)

        # If we have a vm_name, add it to the log message with a white color
        vm_name = ""
        if self._vm_name:
            vm_name = f"{self.BRIGHT_WHITE}*{self._vm_name}*{self.RESET} "

        # Customize log format
        log_format = f"%(asctime)s {vm_name}[{log_color}%(levelname)-8s{self.RESET}] %(message)s"

        # Set the format and return formatted message
        formatter = logging.Formatter(log_format, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


def setup_logging(level: int = logging.INFO, vm_name: str | None = None) -> None:
    """Set up logging."""

    # Setup logger and level
    logger = logging.getLogger()
    logger.setLevel(level)
    # Setup console handler
    console_handler = logging.StreamHandler()
    # Use a better format for messages
    console_handler.setFormatter(IliVMLogFormatter(vm_name=vm_name))
    logger.addHandler(console_handler)
